# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-06-21 16:28
from __future__ import unicode_literals

import datetime
import calendar

import django.contrib.postgres.fields.ranges
from django.db import migrations
from django.core.exceptions import FieldError
from django.db.utils import DataError
from django.conf import settings

import pytz


DEFAULT_PUB_DATE = datetime.datetime(2015, 1, 1)
tz = pytz.timezone(settings.TIME_ZONE)


def end_of_month(dt):
    last_day_number = calendar.monthrange(year=dt.year, month=dt.month)[1]
    last_day = dt.replace(day=last_day_number)
    return last_day + datetime.timedelta(days=1)


def get_date_range(model):
    """Use the existing date fields to infer the start date for our new range
    field"""
    resolution = model.pub_date_resolution

    pub_date = model.pub_date.astimezone(tz)

    if resolution not in ('m', 'w', 'd', 't'):
        raise FieldError(
            'Invalid pub_date_resolution "%s" on model "%s"' % (
                model.pub_date_resolution, model
            ))

    if resolution == 'm':
        start = pub_date.replace(day=1, hour=0, minute=0)
        end = end_of_month(start)
    elif resolution == 'w':
        pub_date = datetime.datetime.combine(
            pub_date,
            datetime.time.min
        ) + datetime.timedelta(days=1)
        end = tz.localize(pub_date)
        start = end - datetime.timedelta(days=7)
    elif resolution == 'd':
        pub_date = datetime.datetime.combine(
            pub_date,
            datetime.time.min
        ) + datetime.timedelta(days=1)
        end = tz.localize(pub_date)
        start = end - datetime.timedelta(hours=24)
    elif resolution == 't':
        start = pub_date
        end = start + datetime.timedelta(minutes=1)

    return (start.replace(second=0, microsecond=0),
            end.replace(second=0, microsecond=0))


def pub_date_to_daterange(apps, schema_editor):
    """Take the stored pub_date and pub_date_resolution fields and turn them
    into a tuple to store in a DateTimeRangeField"""
    Package = apps.get_model("budget", "Package")
    for package in Package.objects.all():
        if package.pub_date is None:
            print("Skipping %s" % package)
            continue

        try:
            package.publish_date = get_date_range(package)
            package.save()
        except DataError as e:
            print('Error processing "%s" with pub_date "%s (%s)".' % (
                package, package.pub_date, package.pub_date_resolution))
            raise e


def default_dateranges(apps, schema_editor):
    """Walk back the date range migration to the field defaults"""
    Package = apps.get_model("budget", "Package")
    for package in Package.objects.all():
        package.publish_date = (DEFAULT_PUB_DATE, DEFAULT_PUB_DATE)
        package.save()


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0007_auto_20160531_1748'),
    ]

    operations = [
        migrations.AddField(
            model_name='package',
            name='publish_date',
            field=django.contrib.postgres.fields.ranges.DateTimeRangeField(default=(DEFAULT_PUB_DATE, DEFAULT_PUB_DATE)),
            preserve_default=False,
        ),
        migrations.RunPython(
            pub_date_to_daterange,
            default_dateranges
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-06-27 22:50
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0016_auto_20160627_1312'),
    ]

    operations = [
        migrations.AlterField(
            model_name='package',
            name='pub_date',
            field=models.DateTimeField(blank=True, editable=False, null=True),
        ),
        migrations.AlterField(
            model_name='package',
            name='pub_date_resolution',
            field=models.CharField(blank=True, choices=[(b'm', b'Month'), (b'w', b'Week'), (b'd', b'Day'), (b't', b'Time')], default=b'd', editable=False, max_length=1, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10b1 on 2016-06-30 22:36
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0020_remove_package_serialized_object'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='package',
            name='print_section',
        ),
        migrations.AddField(
            model_name='package',
            name='print_section',
            field=models.ManyToManyField(blank=True, to='budget.PrintSection'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-04-24 18:43
from __future__ import unicode_literals

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0003_auto_20160420_1800'),
    ]

    operations = [
        migrations.RenameField(
            model_name='package',
            old_name='print_issue',
            new_name='print_run_date',
        ),
        migrations.RemoveField(
            model_name='package',
            name='print_pages',
        ),
        migrations.AddField(
            model_name='package',
            name='is_print_placement_finalized',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='package',
            name='print_placements',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(blank=True, max_length=15, null=True), blank=True, null=True, size=None),
        ),
    ]



# Imports from python.  # NOQA
from datetime import datetime
import hashlib
import json
import time


# Imports from Django.
from django.conf import settings
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.staticfiles.templatetags.staticfiles import static
from django.http import Http404
from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils import timezone
from django.views.generic import TemplateView
from django.views.generic import View


# Imports from other dependencies.
from djangorestframework_camel_case.parser import CamelCaseJSONParser
from djangorestframework_camel_case.render import CamelCaseJSONRenderer
import redis
from rest_framework.authentication import SessionAuthentication
from rest_framework.authentication import TokenAuthentication
from rest_framework.permissions import IsAuthenticated
from rest_framework.renderers import BrowsableAPIRenderer
from rest_framework import viewsets
from six.moves.urllib.parse import urlparse
from six.moves.urllib.parse import urlunparse


# Imports from budget.
from budget.filters import ContentPlacementFilter
from budget.filters import HeadlineViewFilter
from budget.filters import ItemViewFilter
from budget.filters import PackageViewFilter
from budget.filters import PrintPublicationViewFilter
from budget.models import ContentPlacement
from budget.models import Headline
from budget.models import HeadlineVote
from budget.models import Item
from budget.models import Package
from budget.models import PrintPublication
from budget.paginators import ItemViewPagination
from budget.serializers import ContentPlacementSerializer
from budget.serializers import HeadlineSerializer
from budget.serializers import HeadlineVoteSerializer
from budget.serializers import ItemSerializer
from budget.serializers import PackageSerializer
from budget.serializers import PrintPublicationSerializer


REDIS_URL = getattr(settings, 'BUDGET_REDIS_URL', '')

REDIS_CONNECTION = redis.from_url(REDIS_URL)

EDITING_MARKER_TIMEOUT = 90  # Ninety seconds (1.5 minutes).


class MainBudgetView(TemplateView):
    """TK."""
    template_name = 'budget/index.html'

    def get_context_data(self, *args, **kwargs):
        context = super(MainBudgetView, self).get_context_data(*args, **kwargs)

        hostURL = self.request.META['HTTP_HOST']
        aliased_origins = getattr(settings, 'BUDGET_ALIASED_ORIGINS', [])

        context.update({
            'budget_name': getattr(settings, 'BUDGET_TOOL_NAME', 'Budget'),
        })

        # Config AJAX URL.
        config_url = reverse('budget:config')
        if hostURL is not None:
            formatted_url = '{}://{}'.format(self.request.scheme, hostURL)

            if formatted_url in aliased_origins:
                root_budget_url = reverse(
                    'budget:main-budget-view',
                    kwargs={'path': ''}
                )
                config_url = config_url.lstrip(root_budget_url)

                if config_url[0] != '/':
                    config_url = '/{}'.format(config_url)

        context.update({
            'config_url': config_url,
        })

        return context


class ExceptionView(View):
    def get(self, request, *args, **kwargs):
        return 1 / 0


class ConfigView(View):
    def get(self, request, *args, **kwargs):
        originURL = request.GET.get('origin')
        aliased_origins = getattr(settings, 'BUDGET_ALIASED_ORIGINS', [])

        empty_admin_case = [['', 'test@example.com']]
        django_admins = getattr(settings, 'ADMINS', empty_admin_case)
        default_admin_email = django_admins[0][1]

        actual_admin_email = getattr(
            settings,
            'BUDGET_ADMIN_EMAIL',
            default_admin_email
        )

        api_bases = {
            'auth': reverse('apiauth:api-root'),
            'budget': reverse('budget:api-root'),
            'staff': reverse('editorial_staff:api:v1:root'),
        }

        explicit_api_bases = getattr(settings, 'BUDGET_API_CONFIGS', {})

        for key, val in explicit_api_bases.items():
            if callable(val):
                raw_value = val(reverse)
            else:
                raw_value = val

            api_bases[key] = raw_value

        for k, endpoint_url in api_bases.items():
            if originURL is not None and originURL in aliased_origins:
                endpoint_url_parts = urlparse(endpoint_url)

                aliased_url = getattr(settings, 'BUDGET_ALIASED_API_URL', '')
                aliased_url_parts = urlparse(aliased_url)

                final_url = urlunparse(endpoint_url_parts._replace(
                    scheme=aliased_url_parts.scheme,
                    netloc=aliased_url_parts.netloc,
                    path=endpoint_url_parts.path.replace('//', '/')
                ))

                api_bases[k] = final_url

        if originURL is not None and originURL in aliased_origins:
            root_url = None
        else:
            root_url = reverse('budget:main-budget-view', kwargs={'path': ''})

        if request.user.is_authenticated():
            # Use user ID as unique-per-user part of hash.
            unique_user_id = request.user.username
        else:
            # Use user IP address (or HTTP_X_FORWARDED_FOR header, if set) as
            # unique-per-user part of hash.
            x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')

            if x_forwarded_for:
                unique_user_id = x_forwarded_for.split(',')[0]
            else:
                unique_user_id = request.META.get('REMOTE_ADDR')

        unique_pageload_id = hashlib.sha1(
            json.dumps({
                'time_accessed': time.mktime(timezone.now().timetuple()),
                'user': unique_user_id,
            }, sort_keys=True).encode('utf-8')
        ).hexdigest()

        return JsonResponse({
            'adminEmail': actual_admin_email,
            'apiBases': api_bases,
            'branding': {
                'mastheadLogoAltText': getattr(
                    settings,
                    'BUDGET_ORGANIZATION_NAME',
                    'Your organization here'
                ),
                'mastheadLogoURL': static(getattr(
                    settings,
                    'BUDGET_ORGANIZATION_LOGO_PATH',
                    'budget/images/your_name_here.svg'
                )),
                'siteLogoAltText': getattr(
                    settings,
                    'BUDGET_TOOL_NAME',
                    'Budget'
                ),
                'siteLogoURL': static(getattr(
                    settings,
                    'BUDGET_TOOL_LOGO_PATH',
                    'budget/images/budget_logo.svg'
                )),
                'printLogoURL': static(getattr(
                    settings,
                    'BUDGET_PRINT_LOGO_PATH',
                    'budget/images/the-daily.svg'
                )),
            },
            'defaultTimezone': timezone.get_default_timezone_name(),
            'editingURL': reverse('budget:user-editing-package', kwargs={
                'package_id': 0,
            }),
            'exitedURL': reverse('budget:user-exited-package', kwargs={
                'package_id': 0,
            }),
            'externalURLs': json.dumps(
                getattr(settings, 'BUDGET_EXTERNAL_URLS', {})
            ),
            'imageAssets': {
                'notepad': static('budget/images/notepad.svg'),
                'slackProgress': static('budget/images/slack-in-progress.png'),
            },
            'maxItemsToLoad': getattr(settings, 'BUDGET_API_MAX_ITEMS', 500),
            'pageLoadID': unique_pageload_id,
            'printSlugName': getattr(settings, 'BUDGET_PRINT_SLUG_NAME', None),
            'showHeadlines': getattr(settings, 'BUDGET_SHOW_HEADLINES', False),
            'rootURL': root_url,
        })


##########
# Mixins #
##########

class CsrfExemptSessionAuthentication(SessionAuthentication):
    """A custom auth subclass to disable CSRF on the session auth middleware"""
    def enforce_csrf(self, request):
        return


class SessionAndTokenAuthedViewSet(object):
    """Mixin to add both token and Django session auth to a REST viewset"""
    authentication_classes = (CsrfExemptSessionAuthentication,
                              TokenAuthentication,)
    permission_classes = (IsAuthenticated,)


class CamelCasedViewSet(object):
    renderer_classes = (BrowsableAPIRenderer, CamelCaseJSONRenderer,)
    parser_classes = (CamelCaseJSONParser,)


############
# Viewsets #
############


class PackageViewSet(SessionAndTokenAuthedViewSet, CamelCasedViewSet,
                     viewsets.ModelViewSet):
    serializer_class = PackageSerializer
    filter_class = PackageViewFilter

    queryset = Package.objects.distinct().prefetch_related(
        'additional_content',
        'headline_candidates',
        # 'print_section',
    )

    def get_object(self):
        """Attempt to retrieve the package by multiple means.

        If the PK passed is numeric, try getting the package by its ID.
        Otherwise, use the slug instead.
        """
        queryset = self.get_queryset()
        queryset = self.filter_queryset(queryset)

        try:
            int(self.kwargs.get('pk', ''))
            return get_object_or_404(queryset, **self.kwargs)
        except ValueError:
            try:
                return queryset.get_by_slug(self.kwargs['raw_slug'])
            except Package.DoesNotExist:
                raise Http404


class ItemViewSet(SessionAndTokenAuthedViewSet, CamelCasedViewSet,
                  viewsets.ModelViewSet):
    queryset = Item.objects.select_related(
        'primary_for_package',
        'additional_for_package'
    )
    serializer_class = ItemSerializer
    filter_class = ItemViewFilter
    pagination_class = ItemViewPagination


class ContentPlacementViewSet(SessionAndTokenAuthedViewSet, CamelCasedViewSet,
                              viewsets.ModelViewSet):
    serializer_class = ContentPlacementSerializer
    filter_class = ContentPlacementFilter

    queryset = ContentPlacement.objects.distinct()


class HeadlineViewSet(SessionAndTokenAuthedViewSet, CamelCasedViewSet,
                      viewsets.ModelViewSet):
    queryset = Headline.objects.all().order_by('-pk')
    serializer_class = HeadlineSerializer
    filter_class = HeadlineViewFilter


class HeadlineVoteViewSet(SessionAndTokenAuthedViewSet, CamelCasedViewSet,
                          viewsets.ModelViewSet):
    queryset = HeadlineVote.objects.all().order_by('-pk')
    serializer_class = HeadlineVoteSerializer


class PrintPublicationViewSet(SessionAndTokenAuthedViewSet, CamelCasedViewSet,
                              viewsets.ReadOnlyModelViewSet):
    queryset = PrintPublication.objects.all().prefetch_related('sections')
    serializer_class = PrintPublicationSerializer
    filter_class = PrintPublicationViewFilter


########################
# Multi-user warnings. #
########################

class UserPackageBehaviorMixin(LoginRequiredMixin):
    """TK."""
    def dispatch(self, request, package_id, *args, **kwargs):
        # TODO: Decide whether we care about checking whether the package ID
        # actually exists.
        # If not, this saves us a database query — and takes the response times
        # on local dev down from ~35ms to ~25ms.
        # try:
        #     Package.objects.get(id=package_id)
        # except Package.DoesNotExist:
        #     return JsonResponse({
        #         'status': 404,
        #         'message': 'No package found for ID #{}.'.format(package_id)
        #     })

        if request.user.is_authenticated:
            self.package_id = package_id
            self.user_identifier = self.get_user_identifier()
            self.pageload_id = request.GET.get('pageload', None)

            self.current_visit_key = self.get_current_visit_key()

        return super(
            UserPackageBehaviorMixin,
            self
        ).dispatch(request, *args, **kwargs)

    def get_user_identifier(self):
        if self.request.user.email:
            return self.request.user.email
        return self.request.user.username

    def get_current_visit_key(self):
        return 'p={}--u={}--l={}'.format(
            self.package_id,
            self.user_identifier,
            self.pageload_id
        )


class UserEditingPackageUpdateView(UserPackageBehaviorMixin, View):
    def get(self, request, *args, **kwargs):
        # Get the current time.
        # We'll use this for the key value, as well as for the expiration time.
        current_timestamp = int(time.mktime(datetime.now().timetuple()))

        # Calculate new expiration.
        new_expiration_timestamp = current_timestamp + EDITING_MARKER_TIMEOUT

        response_object = {
            'status': 200,
            'current_key': self.current_visit_key,
        }

        # Get or create package key with current timestamp as value.
        old_timestamp_raw = REDIS_CONNECTION.getset(
            self.current_visit_key,
            current_timestamp
        )

        # Set (or increment) the time this key will expire.
        REDIS_CONNECTION.expireat(
            self.current_visit_key,
            new_expiration_timestamp
        )

        if old_timestamp_raw is not None:
            old_timestamp = int(old_timestamp_raw.decode('utf-8'))

            response_object['created'] = False
            response_object[
                'action_message'
            ] = 'Extended active status by {} seconds.'.format(
                current_timestamp - old_timestamp
            )
        else:
            response_object['created'] = True
            response_object[
                'action_message'
            ] = 'New active status until {}.'.format(
                new_expiration_timestamp
            )

        matching_keys = REDIS_CONNECTION.scan(
            match='p={}--*'.format(self.package_id)
        )[1]

        other_active_sessions = [
            key.decode('utf-8') for key in matching_keys
            if key.decode('utf-8') != self.current_visit_key
        ]

        other_active_session_users = list(set([
            key.split('--')[1].split('=')[1] for key in other_active_sessions
        ]))

        response_object['current_user_has_duplicate_sessions'] = (
            self.user_identifier in other_active_session_users
        )

        other_active_users = [
            user_str for user_str in other_active_session_users
            if user_str != self.user_identifier
        ]

        response_object['other_active_users'] = other_active_users

        return JsonResponse(response_object)


class UserExitedPackageUpdateView(UserPackageBehaviorMixin, View):
    def get(self, request, *args, **kwargs):
        # Remove the key.
        REDIS_CONNECTION.delete(self.current_visit_key)

        # Get the same key as was generated above, then delete it.
        return JsonResponse({
            'status': 204,
            'message': 'Successfully deleted key',
            'deleted_key': self.current_visit_key,
        })


# The following code deletes all active editing records for package #114:
# r.delete(*[_.decode('utf-8') for _ in r.scan(match='p=114--*')[1]])



# Imports from python.  # NOQA
from datetime import datetime
import re
import string


# Imports from django.
from django.utils import timezone


# Imports from other dependencies.
from dateutil.relativedelta import relativedelta
from psycopg2.extras import DateTimeTZRange


SLUG_MATCH_RE = re.compile(
    r"""
    ^
    (?P<hub_name>[-\w]+)\.
    (?P<slug_key>[-\w]+)\.
    (?P<month>[\d]{2})(?P<day>([\d]{2}|[-]{2}))(?P<year>[\d]{2})
    (?P<suffix_letter>[\w]{0,1})
    (.(?P<additional_slug>[-\w]+)){0,1}
    $
    """,
    re.VERBOSE
)


def alphacode(i):
    """Return base 26 alpha code.

    e.g., 1 = A, 27 = AA, etc.
    """
    letters = list(string.ascii_lowercase)
    code = ''
    while i > 0:
        i -= 1
        remainder = i % 26
        code = '{}{}'.format(letters[int(remainder)], code)
        i = (i - remainder) / 26
    return code


def slug_date_to_range(raw_date_string):
    """Parse slug's date ('yymmdd' or 'yy--dd') to Date object."""
    is_single_day = True

    raw_month = raw_date_string[0:2]
    raw_day = raw_date_string[2:4]
    raw_year = raw_date_string[4:6]

    if raw_day == '--':
        raw_day = '01'
        is_single_day = False

    first_day = timezone.make_aware(
        datetime.strptime(
            '20{}-{}-{}'.format(raw_year, raw_month, raw_day),
            '%Y-%m-%d'
        )
    )

    delta = relativedelta(days=+1)
    if not is_single_day:
        delta = relativedelta(months=+1)

    return DateTimeTZRange(
        lower=first_day,
        upper=(first_day + delta)
    )



# Imports from python.  # NOQA
from datetime import datetime


# Imports from django.
from django.conf import settings
from django.contrib.postgres.search import SearchQuery
from django.contrib.postgres.search import SearchRank
from django.contrib.postgres.search import SearchVector
from django.db.models import F, Func, Q  # NOQA


# Imports from other dependencies.
from django_filters import BaseInFilter
from django_filters import CharFilter
from django_filters import NumberFilter
from django_filters import OrderingFilter
from psycopg2.extras import DateRange
from psycopg2.extras import DateTimeTZRange
import pytz
from rest_framework import filters


# Imports from budget.
from budget.models import ContentPlacement
from budget.models import Headline
from budget.models import Item
from budget.models import Package
from budget.models import PrintPublication
from budget.utils import SLUG_MATCH_RE


tz = pytz.timezone(settings.TIME_ZONE)


def daterange_filter_decorator(time=True):
    """Decorator that handles datetime-filter validation/parsing logic.

    Enforces the existence of an upper and lower bounds, separated by a
    comma in YYYY-MM-DD format. Returns an empty queryset when
    validation or parsing fails.
    """
    def wrap(filter_func):
        def filter_wrapper(self, queryset, field_name, date_range):
            try:
                lower, upper = date_range.split(',')
            except ValueError:
                return queryset.none()

            try:
                lower = tz.localize(datetime.strptime(
                    lower, '%Y-%m-%d'))
                upper = tz.localize(datetime.strptime(
                    upper, '%Y-%m-%d'))
            except ValueError:
                return queryset.none()

            if lower >= upper:
                return queryset.none()

            if time:
                return filter_func(self, queryset, lower, upper)
            else:
                return filter_func(self, queryset, lower.date(), upper.date())
        return filter_wrapper
    return wrap


class NumberInFilter(BaseInFilter, NumberFilter):
    pass


class PackageOrderingFilter(OrderingFilter):
    def filter(self, qs, ordering_list):
        ordering = ordering_list[0]

        if ordering == 'slug':
            qs = qs.annotate(
                publish_date_upper=Func(
                    F('publish_date'), function='UPPER'),
                publish_date_lower=Func(
                    F('publish_date'), function='LOWER')
            )
            qs = qs.annotate(
                publish_date_len=(
                    F('publish_date_upper') - F('publish_date_lower'))
            )

            qs = qs.order_by(
                '-publish_date_upper',
                'publish_date_len',
                'hub__name',
                'slug_key'
            )

        return qs


class PackageViewFilter(filters.FilterSet):
    id__in = NumberInFilter(name='pk', lookup_expr='in')
    publish_date = CharFilter(method='publish_date_filter')
    content_type = CharFilter(method='content_type_filter')
    person = CharFilter(method='person_filter')
    has_primary = CharFilter(method='has_primary_filter')
    hub = CharFilter(name='hub', lookup_expr='iexact')
    vertical = CharFilter(name='vertical', lookup_expr='iexact')
    search = CharFilter(method='search_filter')
    url_contains = CharFilter(name='published_url', lookup_expr='icontains')
    ordering = PackageOrderingFilter(
        choices=(
            ('slug', 'Slug',),
            # We've retained publish_date as an option for backward
            # compatibility, even though it's now the model manager's default
            # now that django-filter no longer allows a default sort:
            # https://github.com/carltongibson/django-filter/pull/472
            ('publish_date', 'Publish date',),
        )
    )

    def search_filter(self, queryset, field_name, search_term):
        """Base condition (non-faceted) search logic.

        If 'search_term' is a slug (as matched by 'SLUG_MATCH_RE'), look
        for a package with that slug and return it.

        Otherwise, use Django's Postgres full-text search integration to
        search the budget line and slug key fields for matches.
        """
        slug_match = SLUG_MATCH_RE.match(search_term)
        if slug_match:
            try:
                matching_package = queryset.get_by_slug(
                    '{hub_name}.{slug_key}.{month}{day}{year}'.format(
                        **slug_match.groupdict()
                    )
                )
                return queryset.filter(id=matching_package.id)
            except Package.DoesNotExist:
                pass

        vector = SearchVector(
            'primary_content__budget_line', weight='A'
        ) + SearchVector(
            'slug_key',
            'additional_content__budget_line',
            weight='B'
        ) + SearchVector(
            'additional_content__slug_key',
            weight='C'
        )

        return queryset.annotate(
            rank=SearchRank(vector, SearchQuery(search_term))
        ).filter(
            rank__gte=0.1
        ).order_by('-rank')

    def has_primary_filter(self, queryset, field_name, on_or_off):
        if on_or_off not in ('1', '0',):
            return queryset.none()

        filter_val = False if on_or_off == '1' else True
        return queryset.filter(primary_content__isnull=filter_val)

    def person_filter(self, queryset, name, person_email):
        return queryset.filter(
            Q(
                primary_content__editors__contains=[
                    {'email': person_email}
                ]
            ) | Q(
                primary_content__authors__contains=[
                    {'email': person_email}
                ]
            ) | Q(
                additional_content__editors__contains=[
                    {'email': person_email}
                ]
            ) | Q(
                additional_content__authors__contains=[
                    {'email': person_email}
                ]
            )
        )

    def content_type_filter(self, queryset, field, content_type):
        return queryset.filter(
            Q(
                primary_content__type=content_type
            ) | Q(
                additional_content__type=content_type
            )
        )

    @daterange_filter_decorator()
    def publish_date_filter(self, queryset, lower, upper):
        return queryset.filter(
            publish_date__overlap=DateTimeTZRange(
                lower=lower,
                upper=upper
            )
        )

    class Meta:  # NOQA
        model = Package
        fields = []
        order_by = (
            'publish_date',
            'slug',
        )


class ItemViewFilter(filters.FilterSet):
    id__in = NumberInFilter(name='pk', lookup_expr='in')
    person = CharFilter(method='person_filter')
    publish_date = CharFilter(method='publish_date_filter')
    author = CharFilter(method='author_filter')
    editor = CharFilter(method='editor_filter')
    primary_id = NumberFilter(name='primary_for_package__pk')
    additional_id = NumberFilter(name='additional_for_package__pk')

    class Meta:  # NOQA
        model = Item
        fields = []

    """
    We need the below custom filters to traverse into the JSON fields that
    store author/editor information and, in the case of person_filter, to
    allow us to query across both author and editor fields
    """
    def person_filter(self, queryset, field_name, person_email):
        return queryset.filter(
            Q(
                editors__contains=[
                    {'email': person_email}
                ]
            ) | Q(
                authors__contains=[
                    {'email': person_email}
                ]
            )
        )

    def author_filter(self, queryset, field_name, author_email):
        return queryset.filter(authors__contains=[{'email': author_email}])

    def editor_filter(self, queryset, field_name, editor_email):
        return queryset.filter(editors__contains=[{'email': editor_email}])

    """
    To handle date-bound queries, Qs are required to check the publication date
    on any package (primary or additional) connected to our items
    """
    @daterange_filter_decorator()
    def publish_date_filter(self, queryset, lower, upper):
        return queryset.filter(
            Q(primary_for_package__publish_date__overlap=DateTimeTZRange(
                lower=lower, upper=upper)) |
            Q(additional_for_package__publish_date__overlap=DateTimeTZRange(
                lower=lower, upper=upper))
        )


class HeadlineViewFilter(filters.FilterSet):
    id__in = NumberInFilter(name='pk', lookup_expr='in')

    class Meta:  # NOQA
        model = Headline
        fields = []


class PrintPublicationViewFilter(filters.FilterSet):
    name_or_slug = CharFilter(method='name_or_slug_filter')
    publication_active = CharFilter(method='publication_active_filter')

    class Meta:  # NOQA
        model = PrintPublication
        fields = ['name', 'slug']

    def name_or_slug_filter(self, queryset, field_name, queried_value):
        val = queried_value.replace(' and ', ' & ')
        return queryset.filter(Q(name__iexact=val) | Q(slug__iexact=val))

    def publication_active_filter(self, queryset, field_name, on_or_off):
        """Custom zero-and-one implementation of boolean filter.

        This was implemented as a CharFilter because the built-in
        BooleanFilter interprets 3 as 'True' and 2 as 'False', which is
        crazy.

        These custom methods instead interpret 1 as True and 0 as False.
        """
        if on_or_off not in ('1', '0'):
            return queryset.none()

        filter_val = True if on_or_off == '1' else False
        return queryset.filter(is_active=filter_val)


class ContentPlacementOrderingFilter(OrderingFilter):
    def filter(self, qs, ordering_list):
        ordering = ordering_list[0]

        if ordering == 'run_date':
            qs = qs.annotate(
                run_date_upper=Func(F('run_date'), function='UPPER'),
                run_date_lower=Func(F('run_date'), function='LOWER')
            )
            qs = qs.annotate(
                run_date_len=(F('run_date_upper') - F('run_date_lower'))
            )
            qs = qs.order_by('-run_date_upper', 'run_date_len')

        # elif ordering == 'slug':
        #     qs = qs.annotate(
        #         run_date_upper=Func(F('run_date'), function='UPPER'),
        #         run_date_lower=Func(F('run_date'), function='LOWER')
        #     )
        #     qs = qs.annotate(
        #         run_date_len=(F('run_date_upper') - F('run_date_lower'))
        #     )
        #
        #     qs = qs.order_by(
        #         '-run_date_upper',
        #         'run_date_len',
        #         'package__hub__name',
        #         'package__slug_key'
        #     )

        return qs


class ContentPlacementFilter(filters.FilterSet):
    destination = CharFilter(name='destination__slug', lookup_expr='iexact')
    run_date = CharFilter(method='placement_run_date_filter')
    ordering = ContentPlacementOrderingFilter(
        choices=(
            ('run_date', 'Print run date',),
            ('slug', 'Slug',),
        )
    )
    person = CharFilter(method='person_filter')
    vertical = CharFilter(name='package__vertical', lookup_expr='iexact')
    hub = CharFilter(name='package__hub', lookup_expr='iexact')

    def person_filter(self, queryset, name, person_email):
        return queryset.filter(
            Q(
                package__primary_content__editors__contains=[
                    {'email': person_email}
                ]
            ) | Q(
                package__primary_content__authors__contains=[
                    {'email': person_email}
                ]
            ) | Q(
                package__additional_content__editors__contains=[
                    {'email': person_email}
                ]
            ) | Q(
                package__additional_content__authors__contains=[
                    {'email': person_email}
                ]
            )
        )

    @daterange_filter_decorator(time=False)
    def placement_run_date_filter(self, queryset, lower, upper):
        return queryset.filter(
            run_date__overlap=DateRange(
                lower=lower,
                upper=upper
            )
        )

    class Meta:  # NOQA
        model = ContentPlacement
        fields = ('package', 'destination',)



# Imports from python.  # NOQA
from datetime import datetime


# Imports from Django.
from django.core.exceptions import ValidationError
from django.test import SimpleTestCase
from django.test import TestCase
from django.utils.timezone import make_aware


# Imports from other dependencies.
from psycopg2.extras import DateTimeTZRange


# Imports from budget.
from budget.utils import alphacode
from budget.utils import slug_date_to_range
from budget.validators import hub_exists


__all__ = [
    'SlugDateToRangeTestCase',
    'UtilsTestCase',
    'ValidatorTestCase',
]


class SlugDateToRangeTestCase(SimpleTestCase):
    def test_single_day_range(self):
        """Verify 'budget.utils.slug_date_to_range' single-day behavior.

        The method should turn a single-day formatted slug date
        (e.g., '050115') into a 1-day date range.

        The resulting range should also have correct time zones.
        """
        slug_date = '090115'
        actual_range = slug_date_to_range(slug_date)

        intended_range = DateTimeTZRange(
            lower=make_aware(datetime(2015, 9, 1)),
            upper=make_aware(datetime(2015, 9, 2))
        )
        self.assertEqual(intended_range, actual_range)

    def test_multiple_day_range(self):
        """Verify 'budget.utils.slug_date_to_range' multi-day behavior.

        The method should turn a multiple-day formatted slug date
        '05--15') into a week/month date range.

        The resulting range should also have correct time zones.
        """
        # 28-day month
        feb_slug_date = '02--15'
        feb_actual_range = slug_date_to_range(feb_slug_date)

        feb_intended_range = DateTimeTZRange(
            lower=make_aware(datetime(2015, 2, 1)),
            upper=make_aware(datetime(2015, 3, 1))
        )
        self.assertEqual(feb_intended_range, feb_actual_range)

        # 29-day month
        leap_slug_date = '02--16'
        leap_actual_range = slug_date_to_range(leap_slug_date)

        leap_intended_range = DateTimeTZRange(
            lower=make_aware(datetime(2016, 2, 1)),
            upper=make_aware(datetime(2016, 3, 1))
        )
        self.assertEqual(leap_intended_range, leap_actual_range)

        # 30-day month
        sept_slug_date = '09--15'
        sept_actual_range = slug_date_to_range(sept_slug_date)

        sept_intended_range = DateTimeTZRange(
            lower=make_aware(datetime(2015, 9, 1)),
            upper=make_aware(datetime(2015, 10, 1))
        )
        self.assertEqual(sept_intended_range, sept_actual_range)

        # 31-day month
        oct_slug_date = '10--15'
        oct_actual_range = slug_date_to_range(oct_slug_date)

        oct_intended_range = DateTimeTZRange(
            lower=make_aware(datetime(2015, 10, 1)),
            upper=make_aware(datetime(2015, 11, 1))
        )
        self.assertEqual(oct_intended_range, oct_actual_range)


class UtilsTestCase(SimpleTestCase):
    def test_alpha_code_generator(self):
        """Verify 'budget.utils.alphacode' behavior.

        Integers should be transformed into alphas. Ex: 1 = a, 27 = aa
        """
        self.assertEqual(alphacode(1), 'a')
        self.assertEqual(alphacode(26), 'z')
        self.assertEqual(alphacode(27), 'aa')
        self.assertEqual(alphacode(53), 'ba')


class ValidatorTestCase(TestCase):
    def test_hub_exists(self):
        """Verify 'budget.validators.hub_exists' behavior.

        The 'hub_exists' method should throw a ValidationError when
        passed the slug for a nonexistent hub.
        """
        with self.assertRaises(ValidationError):
            hub_exists('not-a-hub')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-08-21 02:32
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0029_alter_slugs'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='package',
            unique_together=set([('hub', 'slug_key', 'publish_date')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-10-28 23:55
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0035_auto_20170924_1624'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contentplacement',
            name='placement_details',
            field=models.CharField(blank=True, help_text='E.G., print page number.', max_length=25, null=True),
        ),
    ]
